#include <raine.h>
#include <stdio.h>
#include "profile.h"
#include "cpumain.h"
#include "mz80.h"
#include "2151intf.h"
#ifdef RAINE_DEBUG
#include "debug.h"
#endif
#include "mz80help.h"

// Number of cycles before reseting the cycles counter and the timers
#define MAX_CYCLES 0x40000000

#define VERBOSE 0

/******************************************************************/
/* This is my attempt to fix music tempo in games using the sound */
/* chip to synchronize.                                           */
/*                                                                */
/* Description of the problem : */
/* The z80 is either using an irq generated by the sound chip or */
/* is simply waiting for the sound chip status to change. These */
/* sound chips are using timers to generate their irq or to adjust */
/* their status, and we can't just reproduce these timers since */
/* the z80 is taking about 5% of the cpu time of a frame (most */
/* of the irq would happen outside of the z80 frame...) */
/* */
/* Solution : I declare my timers as static structs containing the */
/* handler and the number of cycles to be executed by the z80 before */
/* the timer is trigered. */
/* Then a special z80 frame must be executed where only the required */
/* cycles to triger a timer are executed. This slows a little down the */
/* emulation (because the z80 frame is sliced), but it's more acurate */

typedef struct {
  void (*handler)(int);
  int param;
  UINT32 cycles;
  UINT32 id;;
} TimerStruct;

static UINT32 audio_cpu;

#define MAX_TIMERS 20

static TimerStruct timer[MAX_TIMERS];
static int free_timer = 0;
static int timer_id = 0;
static int z80_frame;

void update_timers() {
  // Called after loading a game to try to recalibrate timers...
  /* The best would probably be to save the timer array. The problem is */
  /* how to find the callbacks ??? So for now I prefer to try to guess */
  /* the correct timers... I know it's risky. */
  int n,cpu = audio_cpu & 0xf;
  int elapsed = Z80_context[cpu].dwElapsedTicks;
  for (n=0; n<free_timer; n++)
    if ((timer[n].cycles < elapsed) ||
	(timer[n].cycles > elapsed + z80_frame)) {
      timer[n].cycles = elapsed + 1000;
    }
}

void z80_irq_handler(int irq) {
  if (irq)
    cpu_interrupt(audio_cpu, 0x38);
  else if (z80intPending) {
    // When irq=0, normally the irq line goes low.
    // I am not totally sure it means dropping any pending interrupt...
    z80intPending = 0;
  }
}

void setup_z80_frame(UINT32 cpu, UINT32 cycles) {
  audio_cpu = cpu;
  z80_frame = cycles;
  timer_id = 0;
  // Until now the only game I found to need this is liquidk.
  // So it should be pretty safe to disable them by default.
  // (if they are allowed in a game not supporting them, sound stops).
}

void reset_timers() {
  free_timer = 0;
  timer_id = 0;
}

void *timer_set(double duration, int param, void (*callback)(int))
{
  UINT32 remaining = duration * fps * z80_frame;
  UINT32 elapsed = mz80GetElapsedTicks(0);
  if (remaining < cyclesRemaining) {
    remaining += cyclesRemaining;
  }

  if (free_timer < MAX_TIMERS) {
#if VERBOSE
      fprintf(stderr,"timer_set %g cycles %d fps %d z80_frame %d cyclesremaining %d\n",duration,remaining,fps,z80_frame,cyclesRemaining);
#endif
    timer[free_timer].handler = callback;
    timer[free_timer].param = param;
    timer[free_timer].cycles = elapsed + remaining;
    timer[free_timer].id = timer_id;
    free_timer++;
  } else {
    fprintf(stderr,"free timers overflow !!!\n");
    exit(1);
  }
  return (void*)timer_id++;
}

void timer_remove(void *the_timer) {
  int n;
  UINT32 id = (UINT32)the_timer;
#if VERBOSE
  fprintf(stderr,"CALLED timer_remove\n");
#endif
  for (n=0; n<free_timer; n++) {
    if (timer[n].id == id)
      break;
  }
  if (n<free_timer) { // found ?
    if (n<free_timer-1) // Not the last one ?
      memmove(&timer[n],&timer[n+1],sizeof(TimerStruct)*(free_timer-n-1));
    free_timer--;
  } else {
    fprintf(stderr,"Timer not found %d!\n",id);
    exit(1);
  }
#if VERBOSE
  fprintf(stderr,"after removal free %d\n",free_timer);
#endif
}


void triger_timers() {
  UINT32 elapsed = mz80GetElapsedTicks(0);
  int n;
#if VERBOSE
  int count=0;
#endif
  /* cyclesRemaining is not reseted by mz80 at the end of its frame... */
  /* If we are here, the frame is over. */
  cyclesRemaining=0;

#if VERBOSE
  fprintf(stderr,"elapsed %d free_timer %d\n",elapsed,free_timer);
#endif

#if 0
  for (n=0; n<free_timer; n++) {
    if (timer[n].cycles <= elapsed) { // Trigered !
      count++;
    }
  }
#endif

  for (n=0; n<free_timer; n++) {
    if (timer[n].cycles <= elapsed) { // Trigered !
#if VERBOSE
      fprintf(stderr,"timer %d elapsed %d diff %d free %d\n",n,elapsed,elapsed - timer[n].cycles,free_timer);
#endif
      (*(timer[n].handler))(timer[n].param);
      // Normally, I would call timer_remove here :
      // timer_remove((void*)timer[n].id);
      // but it is rather unefficient since it's looking for a timer
      // when we know whcih one we want to delete...
      if (n<free_timer-1) // not the last timer ?
	memmove(&timer[n],&timer[n+1],sizeof(TimerStruct)*(free_timer-n-1));
      free_timer--;
#if 1
      if (!_z80iff) {
	// Sometimes 2 timers trigger to close to one another and the z80
	// needs time to handle the interrupt.
	// I really wonder how the original hardware handled this.
	// Maybe there was a minimum delay on the line ?
	ExitOnEI = 1; // Exit at the end of the interrupt
	cpu_execute_cycles(audio_cpu, 60000 );
	ExitOnEI = 0;
	dwElapsedTicks = elapsed; // This frame must not count for the timers
      }
#endif
      n--;
    }
  }
  if (elapsed > MAX_CYCLES) { // time to reset the cpu...
    for (n=0; n<free_timer; n++)
      timer[n].cycles -= MAX_CYCLES;
    dwElapsedTicks -= MAX_CYCLES;
  }
#if VERBOSE
  if (count > 1)
    fprintf(stderr,"*** WARNING triger %d\n",count);
#endif
}

int execute_one_z80_audio_frame(UINT32 frame) {
  int n;
  UINT32 elapsed;
  INT32 min_cycles;

  elapsed = mz80GetElapsedTicks(0);
  if (free_timer > 0) {
    min_cycles = timer[0].cycles;
    for (n=1; n<free_timer; n++)
      if (timer[n].cycles < min_cycles)
	min_cycles = timer[n].cycles;
    min_cycles -= elapsed;

    if (min_cycles <= 1000) {
      min_cycles = 1000;
    }
    if (min_cycles > frame) {
      min_cycles = frame;
    }
  } else {
    min_cycles = frame;
  }
#if VERBOSE
  fprintf(stderr,"asked %d\n",min_cycles);
  if (min_cycles < 0) {
    fprintf(stderr,"cycles FIXED to 1000\n");
    min_cycles = 500;
    exit(1);
  }
#endif
  cpu_execute_cycles(audio_cpu, min_cycles );        // Sound Z80
  frame = (mz80GetElapsedTicks(0) - elapsed); // min_cycles;
#if VERBOSE
  if (abs(frame - min_cycles) > 16)
    fprintf(stderr,"diff %d (pc %x)\n",mz80GetElapsedTicks(0) - elapsed - min_cycles,z80pc);
#endif

  triger_timers();
/*  #ifdef RAINE_DEBUG */
/*    print_debug("Z80:%04x\n",z80pc); */
/*  #endif */
  return frame;
}

void execute_z80_audio_frame() {
  INT32 frame = z80_frame;
  switch_cpu(audio_cpu);
  while (frame > 0) {
    frame -= execute_one_z80_audio_frame(frame);
  }
}
